import shelve, base64, os
import requests


image_db = shelve.open('.kooki_images')


def get_image(image_path):
    image_result = ''
    try:
        image_result = image_db[str(image_path)]
    except KeyError:
        try:
            response = requests.get(image_path)
            image_file = response.content
            image_result = convert_image(image_file)
        except requests.exceptions.MissingSchema:
            image_result = get_local_image(image_path)
    return image_result


def get_local_image(image_path):
    with open(image_path, 'rb') as image_file:
        return convert_image(image_file.read())


def convert_image(image_file):
    image_content_base64 = base64.b64encode(image_file)
    return 'data:image/png;base64,' + image_content_base64.decode('utf-8')



from collections import OrderedDict
from kooki.tools import write_file

from .metadata import Metadata, parse_metadata, data_merge
from .template import apply_template, apply_interpreter


def apply_custom(parts, custom):
    print('apply custom')
    result = []
    for index, part in enumerate(parts):
        result.append(custom(part, index))
    return result


def apply_merge(data):
    print('apply merge')
    result = None

    if isinstance(data, list):
        result = ''
        for content in data:
            result += content

    elif isinstance(data, OrderedDict):
        result = ''
        for key, content in OrderedDict(data).items():
            result += content

    else:
        raise Exception('merging bad data type {}'.format(type(data)))

    return result


def apply_split(data, separator):
    print('apply split')
    parts = data.split(separator)
    return parts


def export_to(name, extension, content):
    file_name = '{0}.{1}'.format(name, extension)
    print('export to {}'.format(file_name))
    write_file(file_name, content)


__all__ = [Metadata, apply_template, parse_metadata, apply_interpreter, data_merge, export_to]



import os
from kooki.common import apply_template


def load_extensions(jars, recipe):
    current = os.getcwd()
    extensions = {}
    path = os.path.join(current, 'extensions')
    if os.path.isdir(path):
        load_extensions_in_directory(path, extensions)
    for jar in jars:
        path = os.path.join(jar, recipe, 'extensions')
        if os.path.isdir(path):
            load_extensions_in_directory(path, extensions)
        path = os.path.join(jar, 'extensions')
        if os.path.isdir(path):
            load_extensions_in_directory(path, extensions)
    return extensions


def load_extensions_in_directory(directory, extensions):
    for file_name in os.listdir(directory):
        path = os.path.join(directory, file_name)
        extension_name = os.path.splitext(file_name)[0]
        if extension_name not in extensions:
            extensions[extension_name] = path


def caller(file_full_path, metadata={}):
    def call(*args, **kwargs):
        with open(file_full_path, 'r') as f:
            file_content = f.read()
            metadata.update(**kwargs)
            result = apply_template(file_content, metadata)
            return result
    return call



class ConversionException(Exception):

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


def required_fields():
    return {
        'book': [['author', 'authors'], 'title', 'year', 'publisher'],
        'journal': [['author', 'authors'], 'title', 'journal', 'year'],
        'conference': [['author', 'authors'], 'title', 'booktitle', 'year'],
        'collection': [['author', 'authors'], 'title', 'booktitle', 'year', 'publisher'],
        'masters thesis': [['author', 'authors'], 'title', 'school', 'year'],
        'phd thesis': [['author', 'authors'], 'title', 'school', 'year'],
        'tech report': [['author', 'authors'], 'title', 'year', 'institution', 'number']}


def type_identifiers():
    return {
        'book': '@book',
        'journal': '@article',
        'conference': '@inproceedings',
        'collection': '@incollection',
        'masters thesis': '@mastersthesis',
        'phd thesis': '@phdthesis',
        'tech report': '@techreport'}


def check_required_fields(item):
    typ = item['type']
    try:
        req_fields = required_fields()[typ]
    except KeyError:
        req_fields = []

    for field in req_fields:
        if type(field) is list:
            ok = False
            for option in field:
                if option in item:
                    ok = True
                    break
            if not ok:
                msg = 'Missing required field:'
                for option in field:
                    msg += ' '' + option + '''
                    if (option != field.last):
                        msg += ' or'
                raise ConversionException(msg)
        else:
            if field not in item:
                raise ConversionException('Missing required field: '' + field + ''')


def process_item(key, item, out_file):
    print('processing ' + str(key))

    try:
        typ = item['type']
    except KeyError:
        raise ConversionException('Missing type')

    try:
        out_file.write(type_identifiers()[typ])
    except KeyError:
        raise ConversionException('Unknown type: ' + str(typ))

    out_file.write('{')

    try:
        authors = item['authors']
    except KeyError:
        try:
            authors = [item['author']]
        except KeyError:
            raise ConversionException('Missing author')

    try:
        item['year']
    except KeyError:
        raise ConversionException('Missing year')

    check_required_fields(item)

    print('writing: ' + key)

    out_file.write(key + ' ,\n')

    out_file.write('  author = {')
    for a in range(len(authors)):
        out_file.write(authors[a])
        if a < len(authors) - 1:
            out_file.write(' and ')
        out_file.write('},\n')

    for field in ['type', 'author', 'authors']:
        try:
            del item[field]
        except KeyError:
            pass

    for key, value in item.items():
        out_file.write('  ' + key + ' = {')
        if isinstance(value, list):
            for v in range(len(value)):
                out_file.write(value[v])
                if v < len(value) - 1:
                    out_file.write(' and ')
        else:
            out_file.write(str(value))
            out_file.write('},\n')

    out_file.write('}\n\n')


def yaml_to_bibtex(data):

    output = ''
    items = {}
    for key, item in data.items():
        try:
            if key in items:
                raise ConversionException('Duplicate key: ' + str(key))
            items[key] = item
            process_item(key, item, output)
        except ConversionException as e:
            print('*** Error while processing an item: ' + str(e))



import pretty_output, os, traceback

from collections import OrderedDict

from kooki.jars import search_file, search_file_or_dir, search_recipe, search_jar
from kooki.extension import load_extensions, caller
from kooki.exception import MissingRecipe, MissingTemplate, MissingMetadata
from kooki.exception import MissingContent, InvalidRecipe
from kooki.common import Metadata, parse_metadata


output = []


def output_search_start():
    global output
    output = []


def output_search(path, fullpath):
    if fullpath:
        output.append({'name': path, 'status': '[found]', 'path': fullpath})
    else:
        output.append({'name': path, 'status': ('[missing]', 'red'), 'path': ''})


def output_search_finish():
    pretty_output.infos(output, [('name', 'blue'), ('status', 'green'), ('path', 'cyan')])


def process_document(document_origin):
    document = document_origin.copy()
    recipe = document.recipe
    jars = document.jars

    # jars
    full_path_jars = []
    for jar in document.jars:
        jar_full_path = search_jar(jar)
        full_path_jars.append(jar_full_path)
    document.jars = full_path_jars

    # template
    pretty_output.title_3('template')
    output_search_start()
    file_full_path = search_file(jars, recipe, document.template)
    output_search(document.template, file_full_path)
    if file_full_path:
        with open(file_full_path, 'r') as f:
            file_read = f.read()
        document.template = file_read
        output_search_finish()
    else:
        output_search_finish()
        raise MissingTemplate(jars, recipe, document.template)

    # metadata
    pretty_output.title_3('metadata')
    metadata_full_path = {}
    output_search_start()
    for metadata in document.metadata:
        file_full_path = search_file(jars, recipe, metadata)
        output_search(metadata, file_full_path)
        if file_full_path:
            with open(file_full_path, 'r') as f:
                file_read = f.read()
            metadata_full_path[file_full_path] = file_read
        else:
            output_search_finish()
            raise MissingMetadata(jars, recipe, metadata)
    document.metadata = metadata_full_path
    output_search_finish()

    document.metadata = parse_metadata(document.metadata)

    # thing for more

    def read(document, metadata={}):
        def callback(file_path):
            file_full_path = search_file(document.jars, document.recipe, file_path)
            ret_content = ''
            if file_full_path:
                with open(file_full_path, 'r') as stream:
                    ret_content = stream.read()
            return ret_content
        return callback

    def path(document):
        def callback(file_path):
            file_full_path = search_file_or_dir(document.jars, document.recipe, file_path)
            return file_full_path
        return callback

    others = {
        'read': read(document),
        'path': path(document)}

    document.metadata.update(others)

    # extensions
    pretty_output.title_3('extensions')
    output_search_start()
    extensions_raw = load_extensions(document.jars, document.recipe)
    extensions = Metadata()
    for extension_name, extension_path in extensions_raw.items():

        splitted = extension_name.split('.')
        current = extensions
        for s in splitted[:-1]:
            if s not in current:
                current[s] = Metadata()
            current = current[s]
        metadata = document.metadata.copy()
        current[splitted[-1]] = caller(extension_path, metadata)

        output_search(extension_name, extension_path)
    output_search_finish()

    # content
    pretty_output.title_3('content')
    content_full_path = OrderedDict()
    output_search_start()
    for content in document.content:
        file_full_path = search_file(jars, recipe, content)
        output_search(content, file_full_path)
        if file_full_path:
            with open(file_full_path, 'r') as f:
                file_read = f.read()
            content_full_path[file_full_path] = file_read
        else:
            output_search_finish()
            raise MissingContent(jars, recipe, content)
    document.content = content_full_path
    output_search_finish()

    # recipe
    pretty_output.title_3('recipe')
    recipe_full_path = search_recipe(recipe)

    if not recipe_full_path or not os.path.isdir(recipe_full_path):
        raise MissingRecipe(recipe)

    try:
        import imp
        fp, pathname, description = imp.find_module('recipe', [recipe_full_path])
        recipe_module = imp.load_module('recipe', fp, pathname, description)
        recipe_module.recipe(document, extensions)

    except Exception as e:
        raise InvalidRecipe(recipe, traceback.format_exc())



import subprocess, os, shutil, tempfile

from kooki.tools import write_file
import pretty_output, textwrap


def latex_to_pdf(name, content):

    current = os.getcwd()
    temp_dir = tempfile.mkdtemp()

    os.chdir(temp_dir)

    tex_file_name = '{0}.tex'.format(name)
    pdf_file_name = '{0}.pdf'.format(name.split('/')[-1])
    relative_output_path = '/'.join(name.split('/')[:-1])
    absolute_output_path = os.path.join(current, relative_output_path)

    tex_file_path = os.path.join(temp_dir, tex_file_name)
    pdf_file_path = os.path.join(temp_dir, pdf_file_name)

    if pretty_output._debug_policy:
        write_file(tex_file_name, content)

    write_file(tex_file_path, content)

    command = textwrap.dedent('''\
        xelatex -interaction=nonstopmode -halt-on-error -output-directory={1} {0}\
        ''').format(tex_file_path, temp_dir)

    log_file = os.path.join(temp_dir, 'xelatex.log')
    command_bib = 'bibtex {}'.format(name)

    command_glossary = 'makeglossaries {}'.format(name)

    with open(log_file, 'w') as f:
        print('execute xelatex')
        subprocess.call(command, shell=True, stdout=f)
        subprocess.call(command_bib, shell=True, stdout=f)
        subprocess.call(command_glossary, shell=True, stdout=f)
        subprocess.call(command, shell=True, stdout=f)
        subprocess.call(command, shell=True, stdout=f)

    if pretty_output._debug_policy:
        print('[debug] XeLaTeX output: cat {0}'.format(log_file))

    if os.path.isfile(pdf_file_path):
        if not os.path.isdir(absolute_output_path):
            os.makedirs(absolute_output_path)
        shutil.copy(pdf_file_path, absolute_output_path)
        print('export {} to {}'.format(pdf_file_name, absolute_output_path))
    else:
        print('the file {} is missing'.format(pdf_file_name))

    # shutil.rmtree(temp_dir)
    os.chdir(current)



import argparse, os, pretty_output, traceback, textwrap

from karamel.command import Command

from kooki.exception import KookiException
from kooki.rule import Rule
from kooki.tools import write_file


__command__ = 'new'
__description__ = 'Create a new kooki.yaml file'


class NewCommand(Command):

    def __init__(self):
        super(NewCommand, self).__init__(__command__, __description__)
        self.add_argument('--config-file', default='kooki.yaml')
        self.add_argument('-n', '--name', help='document name')
        self.add_argument('-t', '--template', help='document template')
        self.add_argument('-r', '--recipe', help='document recipe')
        self.add_argument('-j', '--jars', nargs='+', help='documents kooki jars')
        self.add_argument('-m', '--metadata', nargs='+', help='documents metadata sources')
        self.add_argument('-c', '--content', nargs='+', help='documents content sources')

        help_help_message = 'Show this help message and exit.'
        debug_help_message = 'Show information to help debug the bake processing'

        self.add_argument('-h', '--help', action='help', default=argparse.SUPPRESS,
                          help=help_help_message)
        self.add_argument('-d', '--debug', help=debug_help_message, action='store_true')
        self.add_argument('--no-color', help='The output has no color.', action='store_true')
        self.add_argument('--no-output', help='There is no output.', action='store_true')

    def callback(self, args):
        try:
            pretty_output.set_output_policy(not args.no_output)
            pretty_output.set_color_policy(not args.no_color)
            pretty_output.set_debug_policy(args.debug)

            pretty_output.title_1('new')
            self.check_override_config_file(args.config_file)
            rule = self.create_default_rule(args)
            config_file = rule.export()
            pretty_output.title_2('kooki.yaml')
            print(config_file)
            write_file('kooki.yaml', config_file)
            pretty_output.title_2()

        except KookiException as e:
            pretty_output.error_step('Errors')
            pretty_output.error(e)

        except Exception as e:
            pretty_output.error_step('Errors')
            pretty_output.error(traceback.format_exc()[:-1])

    def check_override_config_file(self, config_file):
        if os.path.exists(config_file):
            print('The \'kooki.yaml\' config file already exist.')
            response = input('Overide it [y/n] (y): ')
            if response == 'n':
                raise SystemExit(1)
            elif response == 'y' or response == '':
                pass

    def create_default_rule(self, args):

        rule = Rule()

        default_name = 'index'
        rule.set_name(args.name, default_name)

        default_recipe = 'web'
        rule.set_recipe(args.recipe, default_recipe)

        default_template = 'default.html'
        rule.set_template(args.template, default_template)

        default_jars = ['default']
        rule.set_jars(args.jars, default_jars)

        default_content = textwrap.dedent('''\
            # @document.name

            This is a default document generated by *kooki new*.

            Enjoy using **kooki**.

            *@document.author*
            ''')
        rule.set_content(args.content, default_content)

        default_metadata = textwrap.dedent('''\
            document:
                name: Kooki Document
                author: The developpers''')
        rule.set_metadata(args.metadata, default_metadata)

        return rule



import textwrap


class KookiException(Exception):
    pass


class NoConfigFileFound(KookiException):

    message = textwrap.dedent('''\
        No '{}' file found.
        Create one by executing 'kooki new'.''')

    def __init__(self, config_file_name):
        message = self.message.format(config_file_name)
        super().__init__(message)


class YamlErrorConfigFileParsing(KookiException):

    message = textwrap.dedent('''\
        Yaml error during config file parsing.
        Please check the format of your config file.''')

    def __init__(self, exc):
        message = self.message
        super().__init__(message)


class YamlErrorConfigFileBadType(KookiException):

    message = textwrap.dedent('''\
        The Yaml parsed should be a dict and it is a {}.
        Please check the format of your config file.''')

    def __init__(self, type_found):
        message = self.message.format(type_found)
        super().__init__(message)


class MissingRecipe(KookiException):

    message = textwrap.dedent('''\
        Cannot find recipe '{}'.
        Please check the recipe of your config file.''')

    def __init__(self, recipe):
        message = self.message.format(recipe)
        super().__init__(message)


class InvalidRecipe(KookiException):

    message = textwrap.dedent('''\
        The recipe '{}' seems to be invalid.
        Please check this error with the author of this recipe.
        Give him the following message:
        {}''')

    def __init__(self, recipe, error):
        message = self.message.format(recipe, error)
        super().__init__(message)


class MissingResource(KookiException):

    message = textwrap.dedent('''\
        Cannot find {0} '{1}' in the following directory:
        {2}
        Please check the {0} of your config file.''')

    def __init__(self, jars, recipe, resource_name, resource):
        from kooki.jars import get_search_paths
        directories = get_search_paths(jars, recipe)
        directories_str = '\n'.join(['- {}'.format(directory) for directory in directories])
        message = self.message.format(resource_name, resource, directories_str)
        super().__init__(message)


class MissingTemplate(MissingResource):

    def __init__(self, jars, recipe, template):
        super().__init__(jars, recipe, 'template', template)


class MissingMetadata(MissingResource):

    def __init__(self, jars, recipe, metadata):
        super().__init__(jars, recipe, 'metadata', metadata)


class MissingContent(MissingResource):

    def __init__(self, jars, recipe, content):
        super().__init__(jars, recipe, 'content', content)
